from selenium.webdriver.remote.webdriver import WebDriver

"""
Expected conditions:

To implement your own expected condition, selenium expects from you object with a __call__ implementation.
Its good to remembet that selenium allways pasess WebDriver instance into __call__ method :)
"""

class XpathExists:
    """
    Wait for xpath to exist
    """
    def __init__(self, xpath: str):
        self.__xpath = xpath

    # noinspection PyBroadException
    def __call__(self, driver: WebDriver):
        try:
            driver.find_element_by_xpath(self.__xpath)
            return True
        except:
            return False

    def __str__(self):
        return f'xpath-exists -> {self.__xpath}'

    def __repr__(self):
        return self.__str__()



import logging, sys
from abc import ABC
from abswt.actions import Actions


logger = logging.getLogger('abs-page')
stream_handler = logging.StreamHandler(sys.stdout)
logger.addHandler(stream_handler)


class Page(ABC):
    """
    Page Object / Component base abstraction
    It exposes Actions instance with .actions property

    Example usage:
    
    class SasKodzi(Page):
        url = 'https://sas-kodzi.pl'
        
        BLOG_BUTTON = Locator(Using.XPATH, '//a[@href="/blog"]')

        def goto_posts(self) -> None:
            self.actions.click(self.BLOG_BUTTON.get_by())
            self.actions.wait_for(XpathExists('//body'))
    """

    url = None

    def __init__(self, actions: Actions):
        self.__actions = actions
    
    @property
    def actions(self) -> Actions:
        return self.__actions

    @property
    def title(self):
        return self.__actions.element_provider.driver.title

    def open(self, url: str = None):
        uri = url if url else self.url
        self.actions.goto(uri)



import logging, sys
from abc import ABC, abstractmethod
from enum import Enum
from typing import List, Optional, Tuple

from selenium.webdriver.remote.webelement import WebElement
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.remote.webdriver import WebDriver


class Using(Enum):
    """ Enum for locator types, same as selenium By """
    ID = "id"
    NAME = "name"
    XPATH = "xpath"
    CLASS = "class name"
    CSS = "css selector"
    TAG_NAME = "tag name"
    LINK_TEXT = "link text"
    PARTIAL_LINK_TEXT = "partial link text"


class ParameterExtractor:
    """ Extract {param} parameters from str """
    def __init__(self, text: str) -> None:
        self.__text = text
    
    def get_parameters(self) -> List[str]:
        # todo: refactor ?
        params = []
        param, last_index = self.__find_param(start_index=0)
        while param is not None and last_index > -1:
            params.append(param)
            param, last_index = self.__find_param(start_index=last_index)
        return params
    
    def __find_param(self, start_index: int) -> Tuple[Optional[str], int]:
        # todo: refactor ?
        open = self.__text.find('{', start_index)
        close = self.__text.find('}', start_index + 1)
        if open == -1 or close == -1:
            return None, -1
        if open > close:
            return None, -1
        return self.__text[open + 1:close], close


class Locator:
    """
    Represents WebElement Locator with some extra goodies
    
    Example usage:
    Traditional static locator tuple:
    
    button = Locator(Using.XPATH, '//button')
    # get locator tuple:
    button.get_by()

    
    Goodies:
    Parameterized locator tuple:
    
    parameterized_button = Locator(Using.XPATH, '//button[@name="{button_name}"]')
    foo_by = parameterized_button.get_by(button_name='foo')  # -> ('xpath', '//button[@name="foo"]'
    bar_by = parameterized_button.get_by(button_name='bar')  # -> ('xpath', '//button[@name="bar"]'
    # ! parameters in curly brackets must be passed as keyword arguments into get_by method !
    When forgoten to pass parameters you will get an error:
    parameterized_button.get_by()
    >>> ValueError: get_by method is missing keyword arguments: ['button_name']'
    or
    >>> ValueError: get_by method is missing keyword argument: button_name

    For more examples head to README.md / docs :)
    
    """
    def __init__(self, using: Using, value: str, parameter_extractor: ParameterExtractor = None) -> None:
        self.__using = using
        self.__value = value
        self.__parameter_extractor = parameter_extractor if parameter_extractor else ParameterExtractor(self.__value)
    
    @property
    def using(self) -> Using:
        return self.__using
    
    @property
    def value(self) -> str:
        return self.__value

    @property
    def parameters(self) -> List[str]:
        return self.__parameter_extractor.get_parameters()

    @property
    def is_parameterized(self) -> bool:
        return len(self.parameters) > 0


    def get_by(self, **kwargs) -> Tuple[str, str]:
        """ get (by, value) tuple for finding selenium WebElement object """
        return (self.using.value, self.__get_value(**kwargs))

    def __get_value(self, **kwargs) -> str:
        if self.is_parameterized:
            return self.__parameterize_value(**kwargs)
        return self.value
    
    
    def __parameterize_value(self, **kwargs) -> str:
        if len(kwargs.keys()) == 0: raise ValueError(f'get_by method is missing keyword arguments: {self.parameters}')
        self.__validate_parameters(**kwargs)
        return self.value.format(**kwargs)
    
    def __validate_parameters(self, **kwargs):
        for param in self.parameters:
            if param not in kwargs.keys():
                raise ValueError(f'get_by method is missing keyword argument: {param}')


class Finder(ABC):
    """
    Abstraction for finding WebElements
    
    Basic implementation in abs.elements.FluentFinder
    """

    def __init__(self, webdriver: WebDriver) -> None:
        self.__webdriver = webdriver
        logger = logging.getLogger('abs-finder')
        stream_handler = logging.StreamHandler(sys.stdout)
        logger.addHandler(stream_handler)
        self.logger = logger
        self.logger.setLevel(logging.DEBUG)

    @property
    def webdriver(self) -> WebDriver:
        return self.__webdriver
    
    @abstractmethod
    def find_element(self, locator_tuple: tuple, timeout: int = None, condition: object = None) -> WebElement:
        pass

    @abstractmethod
    def find_elements(self, locator_tuple: tuple, timeout: int = None, condition: object = None) -> List[WebElement]:
        pass


class FluentFinder(Finder):
    """
    Default Finder implementation. Should be usefull for most web testing use cases :)

    Example usage:
    finder = FluenFinder(webdriver, default_timeout = 5)

    default_timeout -> is the default WebDriverWait timeout for finding WebElement -> it can be overriden in both of finder methods
    default_single_condition -> default WebDriverWait condition for finding single WebElement -> it can be overriden in both of finder methods
    default_multi_condition -> same as above, except for finding multiple WebElements
    """
    default_single_condition = EC.presence_of_element_located
    default_multi_condition = EC.presence_of_all_elements_located

    def __init__(self, webdriver: WebDriver, default_timeout: int) -> None:
        self.default_timeout = default_timeout
        super().__init__(webdriver)

    def find_element(self, locator_tuple: tuple, timeout: int = None, condition: object = None) -> WebElement:
        t = timeout if timeout else self.default_timeout
        c = condition if condition else self.default_single_condition
        self.logger.debug(f'find_element: {locator_tuple}, timeout: {t} sec, condition: {c}')
        return WebDriverWait(self.webdriver, t).until(c(locator_tuple))

    def find_elements(self, locator_tuple: tuple, timeout: int = None, condition: object = None) -> WebElement:
        t = timeout if timeout else self.default_timeout
        c = condition if condition else self.default_multi_condition
        self.logger.debug(f'find_elements: {locator_tuple}, timeout: {t} sec, condition: {c}')
        return WebDriverWait(self.webdriver, t).until(c(locator_tuple))



import logging, sys
from time import sleep
from selenium.webdriver.remote.webdriver import WebDriver
from selenium.webdriver.support.wait import WebDriverWait
from abswt.elements import Finder


logger = logging.getLogger('abs-actions')
stream_handler = logging.StreamHandler(sys.stdout)
logger.addHandler(stream_handler)

class Actions:
    """
    Action - responsible for webdriver operations:
    - finding elements
    - clicking, typing, submiting etc...

    Parameters:
    finder:  abs.elements.Finder instance (u can use abs.elements.FluentFinder or implement your own :) )
    wait_for_condition_timeout: default wait for condition timeout when using wait_for method
    wait_between: default delay between action method calls, defaults 0sec

    WebDriver and Finder are accessible with properties .webdriver and .finder

    Action methods are preety self explanotary :)
    Most action methods can override timeout (find WebElement timeout) and condition (expected condition for finding WebElement).
    For readability purpouses they should be used as keyword arguments(timeout, condition)

    Example usage:
    finder = FluenFinder(webdriver, default_timeout=5)
    actions = Actions(finder, wait_for_condition_timeout=10, wait_between=1)

    actions.goto(url)
    actions.click(locator)
    actions.click(locator, timeout=10)
    actions.click(locator, condition=EC.visibility_of_element_located)

    For more practical examples check out README.md / docs

    The actions.<method>(locator_tuple: tuple) -> classic tuple for selenium webriver methods ex: ('xpath', '//div/form') or ('id', 'foobar')

    For helpful and super handy selector tuple implementation check out abs.elements.Locator documentation :) and examples in README.md / docs
    """
    # todo: docstring

    def __init__(self, finder: Finder,  wait_for_condition_timeout: int, wait_between: int = 0) -> None:
        self.wait_between_sec = wait_between
        self.wait_for_condition_timeout = wait_for_condition_timeout
        self.__finder = finder

    @property
    def webdriver(self) -> WebDriver:
        return self.finder.webdriver
    
    @property
    def finder(self) -> Finder:
        return self.__finder
    
    def goto(self, url: str) -> None:
        logger.info(f'goto:: {url}')
        self.webdriver.get(url)
    
    def click(self, locator_tuple: tuple, timeout: int = None, condition: object = None) -> None:
        logger.info(f'click:: {locator_tuple}')
        self.finder.find_element(locator_tuple, timeout, condition).click()
        self.sleep()
        
    def type_text(self, locator_tuple: tuple, text: str, timeout: int = None, condition: object = None) -> None:
        logger.info(f'type text:: {locator_tuple}')
        self.finder.find_element(locator_tuple, timeout, condition).send_keys(text)
        self.sleep()
    
    def submit(self, locator_tuple: tuple = None, timeout: int = None, condition: object = None) -> None:
        lt = locator_tuple if locator_tuple else ('xpath', '//form')
        logger.info(f'submit:: {lt}')
        self.finder.find_element(lt, timeout, condition).submit()
        self.sleep()

    def wait_for(self, condition: object, timeout: int = None) -> None:
        t = timeout if timeout else self.wait_for_condition_timeout
        logger.info(f'wait for:: {condition}, timeout: {t} sec')
        WebDriverWait(self.webdriver, t).until(condition)

    def get_attribute(self, locator_tuple: tuple, attr: str, timeout: int = None, condition: object = None) -> str:
        logger.info(f'get attribute:: {locator_tuple} [{attr}]')
        return self.finder.find_element(locator_tuple, timeout, condition)\
            .get_attribute(attr)

    def get_text(self, locator_tuple: tuple, timeout: int = None, condition: object = None) -> str:
        return self.get_attribute(locator_tuple, 'innerText', timeout, condition)
    
    def execute_js(self, js_script: str) -> str:
        logger.info(f'execute js::\n{js_script}')
        return str(self.webdriver.execute_script(js_script))

    def sleep(self, sec: int = None):
        seconds = sec if sec else self.wait_between_sec
        logger.debug(f'sleep:: {seconds} sec')
        sleep(seconds)



from distutils.core import setup
setup(
  name = 'abs-web-testing',
  packages = ['abswt'],
  version = '0.3',
  license='MIT',
  description = 'Action Based Selenium WebTesting library - selenium in more accessible way :)',
  author = 'Tomasz Majk',
  author_email = 'sas.it.tomasz.majk@gmail.com',
  url = 'https://github.com/bigSAS/abs-web-testing',
  download_url = 'https://github.com/bigSAS/abs-web-testing/archive/v0.3.tar.gz',
  keywords = ['selenium', 'action-based', 'web-testing', 'testing', 'abs-web-testing', 'abs'],
  install_requires=['selenium>=3.141.0'],
  classifiers=[
    'Development Status :: 4 - Beta',
    'Intended Audience :: Developers',
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
  ],
)
